# -----------------------------------------------------------------------------
# $Id: Raw.pm 881 2005-03-29 12:33:43Z topia $
# -----------------------------------------------------------------------------
package Log::Raw;
use strict;
use warnings;
use IO::File;
use File::Spec;
use Tiarra::Encoding;
use base qw(Module);
use Module::Use qw(Tools::DateConvert Log::Writer);
use Tools::DateConvert;
use Log::Writer;
use ControlPort;
use Mask;

sub new {
    my $class = shift;
    my $this = $class->SUPER::new(@_);
    $this->{matching_cache} = {}; # <servername,fname>
    $this->{writer_cache} = {}; # <server,Log::Writer>
    $this->{sync_command} = do {
	my $sync = $this->config->sync;
	if (defined $sync) {
	    uc $sync;
	}
	else {
	    undef;
	}
    };
    $this;
}

sub sync {
    my $this = shift;
    $this->flush_all_file_handles;
    RunLoop->shared->notify_msg("Raw logs synchronized.");
}

sub control_requested {
    my ($this,$request) = @_;
    if ($request->ID eq 'synchronize') {
	$this->sync;
	ControlPort::Reply->new(204,'No Content');
    }
    else {
	die ref($this)." received control request of unsupported ID ".$request->ID."\n";
    }
}

sub message_arrived {
    my ($this,$message,$sender) = @_;

    # syncͭǡ饤ȤäåǤꡢĺΥޥɤsync˰פƤ뤫
    if (defined $this->{sync_command} &&
	$sender->isa('IrcIO::Client') &&
	$message->command eq $this->{sync_command}) {
	# Ƥեflush
	# ¾Υ⥸塼Ʊޥɤsync뤫ΤʤΤǡ
	# do-not-send-to-servers => 1ꤹ뤬
	# åΤ˴Ƥޤʤ
	$this->sync;
	$message->remark('do-not-send-to-servers',1);
	return $message;
    }
    $message;
}

sub message_io_hook {
    my ($this,$message,$io,$type) = @_;

    # break with last
    while (1) {
	last unless $io->server_p;
	last unless Mask::match_deep([Mask::array_or_all(
	    $this->config->command('all'))], $message->command);
	my $msg = $message->clone;
	if ($this->config->resolve_numeric && $message->command =~ /^\d{3}$/) {
	    $msg->command(
		(NumericReply::fetch_name($message->command)||'undef').
		    '('.$message->command.')');
	}
	my $server = $io->network_name;
	my $dirname = $this->_server_match($server);
	if (defined $dirname) {
	    my $prefix  = sprintf '(%s/%s) ', $server, do {
		if ($type eq 'in') {
		    'recv';
		} elsif ($type eq 'out') {
		    'send';
		} else {
		    '----';
		}
	    };

	    my $charset = do {
		if ($msg->have_raw_params) {
		    $msg->encoding_params('binary');
		    'binary';
		} elsif ($io->can('out_encoding')) {
		    $io->out_encoding;
		} else {
		    $this->config->charset;
		}
	    };
	    $this->_write($server, $dirname, $msg->time, $prefix .
			      $msg->serialize($charset));
	}
	last;
    }

    return $message;
}

sub _server_match {
    my ($this,$server) = @_;

    my $cached = $this->{matching_cache}->{$server};
    if (defined $cached) {
	if ($cached eq '') {
	    # cache of not found
	    return undef;
	}
	else {
	    return $cached;
	}
    }

    foreach my $line ($this->config->server('all')) {
	my ($name, $mask) = split /\s+/, $line, 2;
	if (Mask::match($mask,$server)) {
	    # ޥå
	    my $fname_format = $this->config->filename || '%Y.%m.%d.txt';
	    my $fpath_format = $name."/$fname_format";

	    $this->{matching_cache}->{$server} = $fpath_format;
	    return $fpath_format;
	}
    }
    $this->{matching_cache}->{$server} = '';
    undef;
}

sub _write {
    # ꤵ줿ե˥إåդɵ롣
    # ǥ쥯ȥ̾դΥޥִ롣
    my ($this,$channel,$abstract_fpath,$time,$line) = @_;
    my $concrete_fpath = do {
	my $basedir = $this->config->directory;
	if (defined $basedir) {
	    Tools::DateConvert::replace("$basedir/$abstract_fpath", $time);
	}
	else {
	    Tools::DateConvert::replace($abstract_fpath, $time);
	}
    };
    my $header = Tools::DateConvert::replace(
	$this->config->header || '%H:%M',
	$time,
       );
    my $always_flush = do {
	if ($this->config->keep_file_open) {
	    if ($this->config->always_flush) {
		1;
	    } else {
		0;
	    }
	} else {
	    1;
	}
    };
    # եɵ
    my $make_writer = sub {
	Log::Writer->shared_writer->find_object(
	    $concrete_fpath,
	    always_flush => $always_flush,
	    file_mode_oct => $this->config->mode,
	    dir_mode_oct => $this->config->dir_mode,
	   );
    };
    my $writer = sub {
	# åͭ
	if ($this->config->keep_file_open) {
	    # Υͥϥå夵Ƥ뤫
	    my $cached_elem = $this->{writer_cache}->{$channel};
	    if (defined $cached_elem) {
		# å夵줿եѥϺΥեȰפ뤫
		if ($cached_elem->uri eq $concrete_fpath) {
		    # ΥեϥɥѤɤ
		    #print "$concrete_fpath: RECYCLED\n";
		    return $cached_elem;
		}
		else {
		    # ե̾㤦դѤäξ硣
		    # ŤեϥɥĤ롣
		    #print "$concrete_fpath: recached\n";
		    eval {
			$cached_elem->flush;
			$cached_elem->unregister;
		    };
		    # ʥեϥɥ
		    $cached_elem = $make_writer->();
		    if (defined $cached_elem) {
			$cached_elem->register;
		    }
		    return $cached_elem;
		}
	    }
	    else {
		# å夵ƤʤΤǡեϥɥäƥå塣
		#print "$concrete_fpath: *cached*\n";
		my $cached_elem =
		    $this->{writer_cache}->{$channel} =
			$make_writer->();
		if (defined $cached_elem) {
		    $cached_elem->register;
		}
		return $cached_elem;
	    }
	}
	else {
	    # å̵
	    return $make_writer->();
	}
    }->();
    if (defined $writer) {
	$writer->reserve("$header $line\n");
    } else {
	# XXX: do warn with properly frequency
	#RunLoop->shared_loop->notify_warn("can't write to $concrete_fpath: ".
	#				      "$header $line");
    }
}

sub flush_all_file_handles {
    my $this = shift;
    foreach my $cached_elem (values %{$this->{writer_cache}}) {
	eval {
	    $cached_elem->flush;
	};
    }
}

sub destruct {
    my $this = shift;
    # ƤƤLog::WriterĤơåˤ롣
    foreach my $cached_elem (values %{$this->{writer_cache}}) {
	eval {
	    $cached_elem->flush;
	    $cached_elem->unregister;
	};
    }
    %{$this->{writer_cache}} = ();
}

1;

=pod
info: ФȤ̿¸
default: off

# LogϤΥ⥸塼ǤϡʲΤ褦դִԤʤ롣
# %% : %
# %Y : ǯ(4)
# %m : (2)
# %d : (2)
# %H : (2)
# %M : ʬ(2)
# %S : (2)

# ¸ǥ쥯ȥꡣTiarraư֤Хѥ~ϻȤʤ
directory: rawlog

# ƹԤΥإåΥեޥåȡά줿'%H:%M'
header: %H:%M:%S

# ե̾Υեޥåȡά줿'%Y.%m.%d.txt'
filename: %Y-%m-%d.txt

# եΥ⡼(8ʿ)ά줿600
mode: 600

# ǥ쥯ȥΥ⡼(8ʿ)ά줿700
dir-mode: 700

# ȤäƤʸɤ褯狼ʤäȤʸɡά줿utf8
# ֤󤳤λ꤬뤳ȤϤʤȻפޤġġ
charset: jis

# NumericReply ̾褷ɽ(Ȥ dump Ǥ̵ʤޤ)
resolve-numeric: 1

# 륳ޥɤɽޥά줿鵭ϿΥޥɤϿ롣
command: *,-ping,-pong

# ƥե򳫤äѤʤˤ뤫ɤ
# Υץ¿ξ硢ǥޤƸΨɤ¸ޤ
# Ͽ٤ƤΥե򳫤ޤޤˤΤǡ50100Υͥ
# ̡Υե˥褦ʾˤϻȤ٤ǤϤޤ
#  fd դ줿硢饤Ȥ(ޤϥФ)³Ǥʤ
# ʥ⥸塼ɤǤʤǤʤʤɤξɾǽ
# ޤlimit ξܺ٤ˤĤƤ OS ΥɥȤ򻲾ȤƤ
-keep-file-open: 1

# keep-file-open ˳ƹԤȤ flush 뤫ɤ
# open/close ٤ϵˤʤ뤬ϼʤ͸
# keep-file-open ͭǤʤʤ̵뤵(1ˤʤ)ޤ
-always-flush: 0

# keep-file-openͭˤ硢ȯ٤˥եɵΤǤϤʤ
# ʬ̤ίޤäƤ񤭹ޤ롣Τᡢե򳫤Ƥ
# ǶȯϤޤ񤭹ޤƤʤǽ롣
# syncꤹȡ¨¤˥ǥ˽񤭹िΥޥɤɲä롣
# ά줿ϥޥɤɲäʤ
sync: sync

# ƥФꡣ̾ʬϥޥǤ롣
# Ҥ줿ǸΤǡƤΥФ˥ޥå"*"ʤɤϺǸ˽񤫤ʤФʤʤ
# ꤵ줿ǥ쥯ȥ꤬¸ߤʤä顢˺롣
# եޥåȤϼ̤ꡣ
# channel: <ǥ쥯ȥ̾> <̾ޥ>
# :
# filename: %Y-%m-%d.txt
# server: ircnet ircnet
# server: others *
# ǤϡircnetΥircnet/%Y.%m.%d.txtˡ
# ʳΥothers/%Y.%m.%d.txt¸롣
server: ircnet ircnet
server: others *
=cut
